/*
 * LED Matrix display driver (main function and tcpip routines)
 *
 * Controlling several LED Matrix Modules Samsung SLM1606M/SLM1608M
 * connected to port B/D of Ethernut:
 *
 * Printer   Module
 * -----------------
 * D0 (2)    SELECT (CN2-2)
 * D1 (3)    RED    (CN3-2)
 * D2 (4)    GREEN  (CN3-4)
 * D3 (6)    CLOCK  (CN3-6)
 * D4 (8)    BRIGHT (CN3-8)
 * D5 (9)    RESET  (CN3-10)
 * GND (20)  GND    (CN3-3)
 *
 * Written by Thorsten Erdmann 06/2003 (thorsten.erdmann@gmx.de)
 */

/**************************************************************
 **** CAUTION: The module can be destroyed if applied      ****
 ****          static signals, so NEVER interrupt the      ****
 ****          program without switching off module's      ****
 ****          power source first !!!                      ****
 **************************************************************/
 
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <string.h>
#include <stdio.h>
#include <avr/io.h>

//#include <uart.h>
#include "serial.h"

#include "led.h" 
#include "led_asm.h"
#include "parser.h"
#include "dataflash.h"
#include "application.h"
#include "rtc.h"
#include "joystick.h"

const char banner_P[] PROGMEM="\nLED Matrix Display\nType 'help' for a list of available commands\n";

  
/*
 * read a line from a stream with echo and BS support
 */
char *readln(void)
{	
	static char *line=0;
	static u_char x=0;
  char ch;
  if (!line) 
  {
  	line=malloc(40);
  	if (!line) return 0;
    *line=0;
    x=0;
  }  
  if (!(*line)) x=0;
  if (!UartKbhit()) return 0;

  ch=getchar();
  switch (ch)
  {
    case 0x08:
    case 0x7f:
      if (x>0)
      {
        line[--x]=0;
        putchar(ch);
        putchar(' ');
        putchar(ch);
      }
      break;
    default:
      if (ch>=' ')
      {
        if (x<40)
        {
          putchar(ch);
          line[x++]=ch;
          line[x]=0;
        }
      }
  }
  if ((ch=='\n')||(ch=='\r')) return line;
  return 0;
}

/*******************************************************************************
 *
 ******************************************************************************/
/*
void ShowScreen(void)
{
	u_char x,y;
	putchar('\n');
	for (y=0;y<MAXY;y++)
	{
		for (x=0;x<MAXX;x++) putchar('0'+LEDGetPixel(x,y));
		putchar('\n');
	}
}
*/
/*
 * display menu item
 */
void MenuDisplayIcon(u_char nr)
{ 
  LEDClear();
  LEDRectangle((MAXX>>1)-11,0,22,22,RED);
  // draw icon
  if (appl[nr]->icon) LEDBitmap((MAXX-PRG_RDB(appl[nr]->icon))>>1,((20-PRG_RDB(appl[nr]->icon+1))>>1)+1,RED,appl[nr]->icon);
  // draw label
  if (appl[nr]->name) LEDFStr_P(0,22,MAXX,TE_CENTER,RED,appl[nr]->name);  
  // draw left arrow
  if (nr>0)
  {
    LEDHLine(3,8,1,RED);
    LEDHLine(2,9,2,RED);
    LEDHLine(1,10,3,RED);
    LEDHLine(0,11,4,RED);
    LEDHLine(1,12,3,RED);
    LEDHLine(2,13,2,RED);
    LEDHLine(3,14,1,RED);
  }
  // draw right arrow
  if (appl[nr+1])
  {
    LEDHLine(MAXX-4,8,1,RED);
    LEDHLine(MAXX-4,9,2,RED);
    LEDHLine(MAXX-4,10,3,RED);
    LEDHLine(MAXX-4,11,4,RED);
    LEDHLine(MAXX-4,12,3,RED);
    LEDHLine(MAXX-4,13,2,RED);
    LEDHLine(MAXX-4,14,1,RED);
  }  
  LEDDraw();
} 

/*
 * run the main menu
 */ 
void Menu(void)
{
	char *cmd;
	u_char nr=0, result=0;
	// prompt on console
  printf("\n>");
  joystickInit();
	do
	{
		MenuDisplayIcon(nr);
    
    do
    {
		// check for  a command from console
  	  if ((cmd=readln()))
  	  {
  	  	// interpret the command
  	    result=Processline(cmd,stdin);
  	    *cmd=0;
        printf("\n>");
      }
    } while((!keypad) && (result!=255)); // wait for a keypress
    
    // ...else interpret the keys 	
    if ((keypad&BTN_LEFT) && (nr>0)) nr--;
    if ((keypad&BTN_RIGHT) && (appl[nr+1])) nr++;
    if ((keypad&BTN_ACTION) && (appl[nr]->appl_init))
    {
    	while(keypad);
    	joystickInit();
    	LEDClear();
      memset((char *)buf7seg,0,sizeof(buf7seg));    	
    	appl[nr]->appl_init();
    	LEDClear();
      memset((char *)buf7seg,0,sizeof(buf7seg));    
    	LEDDraw();
      printf("\n>");
    }	
   	while(keypad);
  } while (1); // never leave the menu subroutine
}


/*
 * Main application routine. 
 */
int main(void)
{
  
	char err, line[20];
	char hh,mm,ss;

	// set 115200 baud rate for 14,7Mhz CPU clock 
	UartInit(38400,HANDSHAKE_HARDWARE,0);

  puts_P(banner_P);
  puts_P(PSTR("Initialize LED Panel...\n"));
  LEDInit();
  sei();
  
  LEDClear();
  // check dataflash
  printf_P(PSTR("Checking Dataflash... "));
  DF_get_type(line);
  printf_P(PSTR("found %s\n"),line);
  
  // read RTC and initialize random generator
  printf_P(PSTR("Checking RTC... "));
  if ((err=GetTime(&hh,&mm,&ss)))
    printf_P(PSTR("failed, Error code %i\n"),err);
  else
    if ((hh==0) && (mm==0) && (ss==0))
      printf_P(PSTR("clock not set\n"));
    else
    {
    	printf_P(PSTR("%02i:%02i:%02i\n"),hh,mm,ss);
      srand(hh*1000+mm*100+ss);
    }
    
  // wait until no key is pressed
  err=10;
  while ((err>10) && (keypad)) wait(10);
  if (keypad)
  {
  	printf_P(PSTR("ERROR: stuck key\n"));
  	cmdShowkey();
  }
  Menu();  

  return 0;
}
