/*
 * RTC DS1307 driver
 */
#include <avr/twi.h>
#include <avr/interrupt.h>

 
/*
 * Check if year is a leapyear
 */
int isLeapyear (int year)
{								
  return (((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0));
}

/*
 * calculate length of month
 */
int DaysOfMonth(char month, int year)
{
  switch (month)
  {
    case  1:
    case  3:
    case  5:
    case  7:
    case  8:
    case 10:
    case 12: return 31;
    case  2: return 28 + isLeapyear(year);
    case  4:
    case  6:
    case  9:
    case 11: return 30;
    default: return 0;
  }
}

/*
 * calculate Day of Week of a given date
 */
char DayOfWeek(char day, char month, int year)
{
  unsigned int iMonth;
  long dayDiff = year - 1;
  dayDiff += (year - 1) / 4;
  dayDiff -= (year - 1) / 100;
  dayDiff += (year - 1) / 400;
  for (iMonth = 1; iMonth < month; iMonth++)
    dayDiff += DaysOfMonth(iMonth, year);
  dayDiff += day - 1;
  return ((dayDiff) % 7)+1;
}

/*
 * init I2C interface
 */
void i2c_init(void)
{	
  TWSR=BV(TWPS0);  // prescaler 4
  TWBR=0x0f;       // datarate 16
  PORTD|=3;        // activate pullup
  DDRD&=0xfc;      // set I2C bits to input
  wait(10);
}

/*
 * send an I2C start condition
 */
char i2c_start(void)
{
	char err;
  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);    // send start condition 
  while (!(TWCR & (1<<TWINT)));              // wait until ready
  if ((err=(TWSR & 0xF8)) != TW_START)       // check for error
  {
 	  // if error send stop condition to release bus
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
    return err;
  }
  return 0;
}

/*
 * send I2C repeated start condition 
 */
char i2c_rep_start(void)
{
	char err;
  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);    // send start condition
  while (!(TWCR & (1<<TWINT)));              // wait until ready
  if ((err=(TWSR & 0xF8)) != TW_REP_START)   // check for error
  {
 	  // if error send stop condition to release bus
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
    return err;
  }
  return 0;
}

/*
 * send an I2C slave address write
 */
char i2c_sla_w(char addr)
{
	char err;
  TWDR = addr;   										         // set address
  TWCR = (1<<TWINT) | (1<<TWEN);             // start transfer
  while (!(TWCR & (1<<TWINT)));              // wait until ready
  if ((err=(TWSR & 0xF8)) != TW_MT_SLA_ACK)  // check for error
  {
 	  // if error send stop condition to release bus
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
    return err;
  }
  return 0;
}
	 
/*
 * send an I2C slave address read
 */
char i2c_sla_r(char addr)
{
	char err;
  TWDR = addr|1; 										         // set address
  TWCR = (1<<TWINT) | (1<<TWEN);             // start transfer
  while (!(TWCR & (1<<TWINT)));              // wait until ready
  if ((err=(TWSR & 0xF8)) != TW_MR_SLA_ACK)  // check for error
  {
 	  // if error send stop condition to release bus
    TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
    return err;
  }
  return 0;
}
	 
/*
 * send byte to I2C slave
 */
char i2c_data_w(char data)
{
	char err;
  TWDR = data;   						 		             // set data byte
  TWCR = BV(TWINT) | BV(TWEN);              // start transfer
  while (!(TWCR & BV(TWINT)));               // wait until ready
  if ((err=(TWSR & 0xF8)) != TW_MT_DATA_ACK) // check for error
  {
 	  // if error send stop condition to release bus
    TWCR = BV(TWINT) | BV(TWEN) | BV(TWSTO);
    return err;
  }
  return 0;
}

/*
 * read byte from I2C slave
 */
char i2c_data_r(char ack)
{ 
  TWCR = BV(TWINT) | (ack?BV(TWEA):0) | BV(TWEN);    // Send ACK
  while (!(TWCR & BV(TWINT)));               // wait until ready
	return TWDR;	                             // read data
}

/*
 * send I2C stop condition
 */
void i2c_stop(void)
{
  TWCR = BV(TWINT) | BV(TWEN) | BV(TWSTO);
}
	 
/*
 * set RTC time
 */
char SetTime(char hour, char minute, char second)
{
  i2c_init();

	// convert time from integer to BCD
	char h=((  hour/10)<<4)+(hour%10);
	char m=((minute/10)<<4)+(minute%10);
	char s=((second/10)<<4)+(second%10);
	
  if (i2c_start())      return 1; // send start condition
  if (i2c_sla_w(0xd0))  return 2; // send device address write
  if (i2c_data_w(0x00)) return 3; // send register address
  if (i2c_data_w(s))    return 4; // send seconds
  if (i2c_data_w(m))    return 5; // send minutes
  if (i2c_data_w(h))    return 6; // send hours
  i2c_stop();
  return 0;
}

char SetDate(char day, char month, char year)
{
  i2c_init();

	// convert time from integer to BCD
	char d=((  day/10)<<4)+(  day%10);
	char m=((month/10)<<4)+(month%10);
	char y=(( year/10)<<4)+( year%10);
	char wd=DayOfWeek(day, month, 2000+year);
	
  if (i2c_start())      return 1; // send start condition
  if (i2c_sla_w(0xd0))  return 2; // send device address write
  if (i2c_data_w(0x03)) return 3; // send register address
  if (i2c_data_w(wd))   return 4; // send day of week
  if (i2c_data_w(d))    return 5; // send day
  if (i2c_data_w(m))    return 6; // send month
  if (i2c_data_w(y))    return 7; // send year
  i2c_stop();

  return 0;
}

char GetTime(char *hour, char *minute, char *seconds)
{
	char h,m,s;

  i2c_init();
  if (i2c_start())      return 1; // send start condition
  if (i2c_sla_w(0xd0))  return 2; // send device address write
  if (i2c_data_w(0x00)) return 3; // send register address
  if (i2c_rep_start())  return 4; // send repeated start condition
  if (i2c_sla_r(0xd0))  return 5; // send device address read
  s=i2c_data_r(1)&0x7f;           // read seconds
  m=i2c_data_r(1)&0x7f;           // read minutes
  h=i2c_data_r(0)&0x3f;           // read hours
  i2c_stop();
	*hour   =(10*(h&0xf0)>>4)+(h&0x0f);
	*minute =(10*(m&0xf0)>>4)+(m&0x0f);
	*seconds=(10*(s&0xf0)>>4)+(s&0x0f);
	return 0;
}

char GetDate(char *dw, char *day, char *month, char *year)
{
	char d,m,y;

  i2c_init();
  if (i2c_start())      return 1; // send start condition
  if (i2c_sla_w(0xd0))  return 2; // send device address write
  if (i2c_data_w(0x03)) return 3; // send register address
  if (i2c_rep_start())  return 4; // send repeated start condition
  if (i2c_sla_r(0xd0))  return 5; // send device address read
  *dw=i2c_data_r(1)&0x07;         // read day of week
  d=i2c_data_r(1)&0x3f;           // read day
  m=i2c_data_r(1)&0x1f;           // read month
  y=i2c_data_r(0);                // read year
  i2c_stop();

	*year  =(10*(y&0xf0)>>4)+(y&0x0f);
	*month =(10*(m&0xf0)>>4)+(m&0x0f);
	*day   =(10*(d&0xf0)>>4)+(d&0x0f);
	return 0;
}
